package buffer;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.atomic.AtomicInteger;
import junit.framework.TestCase;

public class PutTakeTest extends TestCase {
	private static final ExecutorService pool = Executors.newCachedThreadPool();
	private final AtomicInteger putSum = new AtomicInteger(0);
	private final AtomicInteger takeSum = new AtomicInteger(0);
	private final CyclicBarrier barrier;
	private final BoundedBuffer<Integer> bb;
	private final int nTrials, nPairs;
	
	public static void main(String[] args) {
		new PutTakeTest(10, 10, 1000).test();
		pool.shutdown();
	}
	public PutTakeTest() {
		this.bb = new BoundedBuffer<Integer>(10);
		this.nTrials = 10;
		this.nPairs = 1000;
		this.barrier = new CyclicBarrier(nPairs*2 + 1);
	}
	
	public PutTakeTest(int capacity, int npairs, int ntrials){
		this.bb = new BoundedBuffer<Integer>(capacity);
		this.nTrials = ntrials;
		this.nPairs = npairs;
		this.barrier = new CyclicBarrier(npairs*2 + 1);
	}
	
	public void setUp() {
		new PutTakeTest(10, 10, 1000);
	}
	public void tearDown() {
		pool.shutdown();
	}
	public void test() {
		try {
			for(int i = 0; i < nPairs; i++){
				pool.execute(new Producer());
				pool.execute(new Consumer());
			}
			barrier.await(); // wait for threads to be ready
			barrier.await(); // wait for threads to be finished
			assertEquals(putSum.get(), takeSum.get());
		}catch(Exception e){
			throw new RuntimeException(e);
		}
	}
	public static int xorShift(int y){
		y ^= (y << 6);
		y ^= (y >>> 21);
		y ^= (y << 7);
		return y;
	}
	
	class Producer implements Runnable {
		public void run() {
			try {
				int seed = (this.hashCode() ^ (int)System.nanoTime());
				int sum = 0;
				barrier.await();
				for(int i=nTrials; i>0; --i){
					bb.put(seed);
					sum += seed;
					seed = xorShift(seed);
				}
				putSum.getAndAdd(sum);
				barrier.await();
			}catch(Exception e){
				throw new RuntimeException(e);
			}
		}
	}
	
	class Consumer implements Runnable {
		public void run() {
			try {
				barrier.await();
				int sum = 0;
				for(int i=nTrials; i> 0; --i){
					sum += bb.take();
				}
				takeSum.getAndAdd(sum);
				barrier.await();
			}catch(Exception e){
				throw new RuntimeException(e);
			}
		}
	}
}
