package buffer;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.atomic.AtomicInteger;

import buffer.PutTakeTest.Consumer;
import buffer.PutTakeTest.Producer;
import junit.framework.TestCase;

public class TimedPutTakeTest extends TestCase {
	private static final ExecutorService pool = Executors.newCachedThreadPool();
	private final AtomicInteger putSum = new AtomicInteger(0);
	private final AtomicInteger takeSum = new AtomicInteger(0);
	private final CyclicBarrier barrier;
	private final BarrierTimer timer;
	private final BoundedBuffer<Integer> bb;
	private final int nTrials, nPairs;
	
	public static void main(String[] args) {
		int tpt = 100000;
		try {
		for(int cap = 1; cap <= 1000; cap *= 10){
			System.out.println("Capacity: " + cap);
			for(int pairs = 1; pairs <= 16; pairs *=2){
				TimedPutTakeTest t = new TimedPutTakeTest(cap, pairs, tpt);
				System.out.println("Pairs:" + pairs + "\t");
				t.test();
				System.out.println("test 1 finished \t");
				Thread.sleep(100);
				t.test();
				System.out.println("test 2 finished");
				Thread.sleep(100);
			}
		}
		}catch(Exception e){e.printStackTrace();}
		pool.shutdown();
	}
	public TimedPutTakeTest() {
		this.bb = new BoundedBuffer<Integer>(10);
		this.nTrials = 10;
		this.nPairs = 1000;
		this.timer = new BarrierTimer();
		this.barrier = new CyclicBarrier(nPairs*2 + 1, timer);
	}
	
	public TimedPutTakeTest(int capacity, int npairs, int ntrials){
		this.bb = new BoundedBuffer<Integer>(capacity);
		this.nTrials = ntrials;
		this.nPairs = npairs;
		this.timer = new BarrierTimer();
		this.barrier = new CyclicBarrier(npairs*2 + 1, timer);
	}
	
	public void test() {
		try {
			timer.clear();
			for(int i = 0; i < nPairs; i++){
				System.out.println("starting thread " + i);
				pool.execute(new Producer());
				pool.execute(new Consumer());
			}
			barrier.await(); // wait for threads to be ready
			barrier.await(); // wait for threads to be finished
			long nsTime = timer.getTime();
			long numItems = nPairs*(long)nTrials;
			long nsPerItem = nsTime/numItems;
			double sTime = nsTime*0.000000001;
			System.out.println("Time passed: " + sTime + " sec, " + numItems + " items");
			System.out.println("Throughput: " + nsPerItem + " ns/item");
			assertEquals(putSum.get(), takeSum.get());
		}catch(Exception e){
			throw new RuntimeException(e);
		}
	}
	public static int xorShift(int y){
		y ^= (y << 6);
		y ^= (y >>> 21);
		y ^= (y << 7);
		return y;
	}
	
	class Producer implements Runnable {
		public void run() {
			try {
				int seed = (this.hashCode() ^ (int)System.nanoTime());
				int sum = 0;
				barrier.await();
				for(int i=nTrials; i>0; --i){
					bb.put(seed);
					sum += seed;
					seed = xorShift(seed);
				}
				putSum.getAndAdd(sum);
				barrier.await();
			}catch(Exception e){
				throw new RuntimeException(e);
			}
		}
	}
	
	class Consumer implements Runnable {
		public void run() {
			try {
				barrier.await();
				int sum = 0;
				for(int i=nTrials; i> 0; --i){
					sum += bb.take();
				}
				takeSum.getAndAdd(sum);
				barrier.await();
			}catch(Exception e){
				throw new RuntimeException(e);
			}
		}
	}
}
